/*
 * mmu.h - memory management routines
 *
 * This file is a part of the multithreading version of the
 * SimpleScalar tool suite written by Todd M. Austin as a part
 * of the Multiscalar Research Project.
 * 
 * Copyright (C) 1997 by Dominik Madon
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dominik.madon@epfl.ch
 * Snail: LSL-DI/EPFL, In-Ecublens, 1015 Lausanne
 *
 * $Id: mmu.h,v 1.6 1998/12/05 00:46:56 madon Exp $
 *
 *
 */


#ifndef MMU_H
#define MMU_H

#include "machine.h"


/* Size of the hash table for address translation */
#define MMU_HASH_SIZE           512

/* Macro to generate the hash key */
#define MMU_HASH_KEY(address)   (MEM_BLOCK(address) % MMU_HASH_SIZE)



/* Type of an entry of the list of addresses */
typedef struct mmu_entry_t {
  md_addr_t logical;
  ph_addr_t physical;
  struct mmu_entry_t *next;
} MMU_ENTRY_TYPE;

/* Type of the MMU database */
typedef struct mmu_db_t {
  MMU_ENTRY_TYPE **thread_context[SS_NUM_CONTEXT];
  unsigned int config[SS_NUM_CONTEXT];
} MMU_DB_TYPE;

/* configuration for each context 
 
    31                 16 15       8 7        0
   +----------+----------+----------+----------+
   |                     | memory context bank |
   +----------+----------+----------+----------+

    memory context bank: Usually set to context number
                         so each context access its own memory bank.

*/





/*                                                                      */
/* m m u _ r e g _ s t a t s                                            */
/*                                                                      */
/* Not implemented yet.                                                 */
/*                                                                      */
/*                                                                      */

/* void mmu_reg_stats(struct stat_sdb_t *sdb); */

/* m m u _ d u m p                                                      */
/*                                                                      */
/* For test purposes only.                                              */
/*                                                                      */
/*                                                                      */

void mmu_dump(MMU_DB_TYPE *mmu_db, FILE *stream);

/*                                                                      */
/* m m u _ i n i t                                                      */
/*                                                                      */
/* Initialize the Memory Managment Unit.                                */
/*                                                                      */
/* Returns a pointer to the mmu.                                        */
/*                                                                      */

MMU_DB_TYPE * mmu_init(void);

/*                                                                      */
/* m m u _ f r e e                                                      */
/*                                                                      */
/* Destroy the mmu and frees the allocated memory.                      */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   inout: the mmu.                                                    */
/*                                                                      */

void mmu_free(MMU_DB_TYPE *mmu_db);

/*                                                                      */
/* m m u _ s e t _ c o n t e x t _ c o n f i g                          */
/*                                                                      */
/* Configure the mmu internal context registers                         */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: thread context.                                                */
/*   in: the new config.                                                */
/*   inout: the mmu.                                                    */
/*                                                                      */

void mmu_set_context_config(
   SS_CONTEXT_TYPE context,
   unsigned int config,
   MMU_DB_TYPE *mmu_db);

/*                                                                      */
/* m m u _ g e t _ c o n t e x t _ c o n f i g                          */
/*                                                                      */
/* Get the value of a context registers configuration                   */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: thread context.                                                */
/*   inout: the mmu.                                                    */
/*                                                                      */
/*   returned: the config.                                              */
/*                                                                      */

unsigned int mmu_get_context_config(
   SS_CONTEXT_TYPE context,
   MMU_DB_TYPE *mmu_db);

/*                                                                      */
/* m m u _ m a p                                                        */
/*                                                                      */
/* Returns the physical address corresponding to the logical given.     */
/* If the page is not in the MMU, this function returns 0 else 1.       */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: thread context.                                                */
/*   in: a logical address.                                             */
/*   inout: the mmu.                                                    */
/*                                                                      */

ph_addr_t mmu_map(
  SS_CONTEXT_TYPE context,
  md_addr_t logical_addr,
  MMU_DB_TYPE *mmu_db);

/*                                                                      */
/* m m u _ i s _ m a p p e d _ b y                                      */
/*                                                                      */
/* Returns the logical address corresponding to the physical given.     */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: thread context.                                                */
/*   in: a physical address.                                            */
/*   inout: the mmu.                                                    */
/*                                                                      */

md_addr_t mmu_is_mapped_by(
  SS_CONTEXT_TYPE context,
  ph_addr_t physical_addr,
  MMU_DB_TYPE *mmu_db);

/* m m u _ e x i s t s _ p a g e                                        */
/*                                                                      */
/* Returns TRUE if the specified exists in the MMU else FLASE.          */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: thread context.                                                */
/*   in: a logical address.                                             */
/*   inout: the mmu.                                                    */
/*                                                                      */

unsigned int mmu_exists_page(
  SS_CONTEXT_TYPE context,
  md_addr_t logical_addr,
  MMU_DB_TYPE *mmu_db);

/*                                                                      */
/* m m u _ a d d _ e n t r y                                            */
/*                                                                      */
/* Add the couple (logical address, physical address) to the            */
/* memory manager unit.                                                 */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: thread context.                                                */
/*   in: logical address.                                               */
/*   in: physical address.                                              */
/*   inout: the mmu.                                                    */
/*                                                                      */

int mmu_add_entry(
  SS_CONTEXT_TYPE context,
  md_addr_t logical_addr,
  ph_addr_t physical_addr,
  MMU_DB_TYPE *mmu_db);


/*                                                                      */
/* m m u _ i n v a l i d a t e  _ e n t r y                             */
/*                                                                      */
/* Suppress the entry given. Depending on the first parameter, it is    */
/* possible to invalidate a full context, a specific entry or a         */
/* combination of them. The target is specified with MMU_INV_*          */
/* constants. The unused parameter can be just NULL.                    */
/*                                                                      */
/* Params:                                                              */
/*                                                                      */
/*   in: invalidation type:                                             */
#define MMU_INV_F_LOGICAL   0x00000001
#define MMU_INV_F_PHYSICAL  0x00000002
#define MMU_INV_F_CONTEXT   0x00000004

/*   in: thread context.                                                */
/*   in: logical address.                                               */
/*   in: physical address.                                              */
/*   inout: the mmu.                                                    */
/*                                                                      */

int mmu_invalidate_entry(
  unsigned int command,                       
  SS_CONTEXT_TYPE context,
  md_addr_t logical_addr,
  ph_addr_t physical_addr,
  MMU_DB_TYPE *mmu_db);

 


#endif /* MMU_H */
